package com.wechi.adweb.bridge.google.analytics;

import static com.wechi.adweb.bridge.google.analytics.dto.report.GAReportRequestDTO.*;

import com.google.analytics.data.v1beta.*;
import com.google.api.gax.core.FixedCredentialsProvider;
import com.google.auth.oauth2.GoogleCredentials;
import com.google.gson.Gson;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.analytics.dto.report.GAReportRequestDTO;
import com.wechi.adweb.bridge.google.analytics.dto.report.ReportType;

import jakarta.annotation.PostConstruct;

import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.io.IOException;
import java.util.List;
import java.util.Objects;

/**
 * @author wfansh
 *     <p>See https://developers.google.com/analytics/devguides/reporting/data/v1/api-schema for GA
 *     metrics and dimensions.
 */
@Slf4j
@Service
public class GADataService {

    @Value("${spring.application.name}")
    private String applicationName;

    @Value("${google.service.account.key}")
    private String serviceAccountKey;

    private BetaAnalyticsDataSettings dataSettings;

    private Gson gson = new Gson();

    @PostConstruct
    private void init() throws IOException {
        GoogleCredentials credentials =
                GoogleCredentials.fromStream(
                        this.getClass().getClassLoader().getResourceAsStream(serviceAccountKey));

        this.dataSettings =
                BetaAnalyticsDataSettings.newBuilder()
                        .setCredentialsProvider(FixedCredentialsProvider.create(credentials))
                        .build();
    }

    public List<Object> runReport(GAReportRequestDTO reportRequest) throws DataException {
        // 0. Enhances the report request.
        this.amplifyReportRequest(reportRequest);

        try (BetaAnalyticsDataClient analyticsDataClient =
                BetaAnalyticsDataClient.create(dataSettings)) {
            // 1. Dimensions.
            List<Dimension> dimensions =
                    reportRequest.getDimensions().stream()
                            .filter(StringUtils::isNotEmpty)
                            .map(dimension -> Dimension.newBuilder().setName(dimension).build())
                            .toList();

            // 2. Metrics.
            List<Metric> metrics =
                    reportRequest.getMetrics().stream()
                            .filter(StringUtils::isNotEmpty)
                            .map(metric -> Metric.newBuilder().setName(metric).build())
                            .toList();

            // 3. Date range.
            DateRange dateRange =
                    DateRange.newBuilder()
                            .setStartDate(reportRequest.getStartDate())
                            .setEndDate(reportRequest.getEndDate())
                            .build();

            // 4. Order by.
            OrderBy.Builder orderBy = null;
            String orderByName = reportRequest.getOrderBy();
            OrderByType orderByType = reportRequest.getOrderByType();
            if (StringUtils.isNotEmpty(orderByName) && Objects.nonNull(orderByType)) {
                orderBy = OrderBy.newBuilder();
                switch (orderByType) {
                    case DIMENSIONS ->
                            orderBy.setDimension(
                                    OrderBy.DimensionOrderBy.newBuilder()
                                            .setDimensionName(orderByName));
                    case METRICS ->
                            orderBy.setMetric(
                                    OrderBy.MetricOrderBy.newBuilder().setMetricName(orderByName));
                }

                // Desc.
                orderBy.setDesc(reportRequest.getOrderByDesc());
            }

            RunReportRequest.Builder request =
                    RunReportRequest.newBuilder()
                            .setProperty(reportRequest.getPropertyName())
                            .addAllDimensions(dimensions)
                            .addAllMetrics(metrics)
                            .addDateRanges(dateRange)
                            // The API returns a maximum of 250,000 rows per request. See
                            // https://developers.google.com/analytics/devguides/reporting/data/v1/rest/v1beta/properties/runReport
                            .setLimit(250_000L);
            if (Objects.nonNull(orderBy)) {
                request.addOrderBys(orderBy);
            }

            RunReportResponse response = analyticsDataClient.runReport(request.build());

            // TODO: Parses the response.
            return null;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    /** Enhances the report request by applying default settings from {@link ReportType}. */
    private void amplifyReportRequest(GAReportRequestDTO reportRequest) {
        ReportType reportType = reportRequest.getReportType();
        if (Objects.isNull(reportType)) {
            return;
        }

        if (CollectionUtils.isEmpty(reportRequest.getMetrics())) {
            reportRequest.setMetrics(reportType.getDefaultMetrics());
        }

        if (CollectionUtils.isEmpty(reportRequest.getDimensions())) {
            reportRequest.setDimensions(reportType.getDefaultDimensions());
        }

        if (Objects.isNull(reportRequest.getOrderBy())) {
            reportRequest.setOrderBy(reportType.getDefaultOrderBy());
        }

        if (Objects.isNull(reportRequest.getOrderByType())) {
            reportRequest.setOrderByType(reportType.getDefaultOrderByType());
        }

        if (Objects.isNull(reportRequest.getOrderByDesc())) {
            reportRequest.setOrderByDesc(reportType.isDefaultOrderByDesc());
        }
    }
}
